/*
 * Copyright (c) 2009, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef INTERACTIVE_MANIPULATION_FRONTEND
#define INTERACTIVE_MANIPULATION_FRONTEND

#include <ros/ros.h>

#include <actionlib/client/simple_action_client.h>

#include <std_msgs/String.h>

#include "pr2_interactive_manipulation/interactive_manipulation_ui.h"
#include "pr2_interactive_manipulation/advanced_options_dialog.h"
#include "pr2_object_manipulation_msgs/IMGUIOptions.h"
#include "pr2_object_manipulation_msgs/IMGUIAdvancedOptions.h"
#include "pr2_object_manipulation_msgs/IMGUIAction.h"
#include "pr2_object_manipulation_msgs/ActionInfo.h"

namespace rviz {
class VisualizationManager;
class FrameManager;
}

namespace pr2_interactive_manipulation {

class InteractiveManipulationFrontend : public InteractiveManipulationFrameBase
{
public:
  InteractiveManipulationFrontend(wxWindow *parent, rviz::VisualizationManager* manager);

  ~InteractiveManipulationFrontend();

  void update();

  void setAdvancedOptions(pr2_object_manipulation_msgs::IMGUIAdvancedOptions ago){adv_options_ = ago;}

  int interface_number_;

  int task_number_;

protected:

  void feedbackCallback(const pr2_object_manipulation_msgs::IMGUIFeedbackConstPtr &feedback);

  void statusCallback( const std_msgs::StringConstPtr &status);

  void rcommanderRefresh();

  virtual void graspButtonClicked( wxCommandEvent& event );

  virtual void placeButtonClicked( wxCommandEvent& event );

  virtual void plannedMoveButtonClicked( wxCommandEvent& event );

  virtual void cancelButtonClicked( wxCommandEvent& event );

  virtual void stopNavButtonClicked( wxCommandEvent& event );

  virtual void resetButtonClicked( wxCommandEvent& event );

  virtual void armGoButtonClicked( wxCommandEvent& event );

  virtual void cameraLeftButtonClicked( wxCommandEvent& event );

  virtual void cameraRightButtonClicked( wxCommandEvent& event );

  virtual void cameraFrontButtonClicked( wxCommandEvent& event );

  virtual void cameraTopButtonClicked( wxCommandEvent& event );

  virtual void advancedOptionsClicked( wxCommandEvent& event );

  virtual void modelObjectClicked( wxCommandEvent& event);

  //virtual void openDrawerButtonClicked( wxCommandEvent& event);

  virtual void rcommandRunButtonClicked( wxCommandEvent& event);

  virtual void rcommandRefreshButtonClicked(wxCommandEvent& event);

  virtual void gripperSliderScrollChanged( wxScrollEvent& );
 
  pr2_object_manipulation_msgs::IMGUIOptions getDialogOptions();

  void setCamera(std::vector<double> params);

  rviz::VisualizationManager* vis_manager_;

  ros::NodeHandle root_nh_;
  ros::NodeHandle priv_nh_;

  ros::ServiceClient rcommander_action_info_client_;
  actionlib::SimpleActionClient<pr2_object_manipulation_msgs::IMGUIAction> *action_client_;
  //actionlib::SimpleActionClient<RunScriptAction> *scripted_action_client_;
  //std::string scripted_action_name_;
  ros::Subscriber status_sub_;
  
  std::string rcommander_action_info_name_;
  std::string rcommander_group_name_;
  std::string action_name_;
  std::string status_name_;

  std::string status_label_text_;

  boost::mutex status_label_mutex_;

  pr2_object_manipulation_msgs::IMGUIAdvancedOptions adv_options_;
};

}

#endif
