/*********************************************************************
*
*  Copyright (c) 2009, Willow Garage, Inc.
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Willow Garage nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

// Author(s): Marius Muja and Matei Ciocarlie

#include "tabletop_vfh_cluster_detector/marker_generator.h"

//for random colors
#include <stdlib.h>
#include <time.h>

#include <geometry_msgs/Pose.h>

#include <tf/transform_datatypes.h>

namespace tabletop_vfh_cluster_detector {

visualization_msgs::Marker MarkerGenerator::getFitMarker(const arm_navigation_msgs::Shape &mesh, double rank)
{
  //create the marker
  visualization_msgs::Marker marker;
  marker.action = visualization_msgs::Marker::ADD;
  marker.lifetime = ros::Duration();
  marker.type = visualization_msgs::Marker::POINTS;
  //marker will be published in object frame, so its pose is identity
  marker.color.a = 1.0;

  marker.color.r = rank;
  marker.color.g = 1 - rank;
  marker.color.b = 0.0;

  marker.scale.x = 0.003;
  marker.scale.y = 0.003;
  marker.points.insert(marker.points.begin(), mesh.vertices.begin(), mesh.vertices.end());

  for (int i=0; i<40; i++)
  {
    geometry_msgs::Point p;
    p.x = 0.0005 * i;
    p.y = p.z = 0.0;
    marker.points.push_back(p);
  }

  return marker;
}

} //namespace tabletop_object_detector
