import vtk
#from Numeric import *
#from numpy import *
#from numpy.linalg import *
import numpy as np
import pickle
from tesselatedSphere import TesselatedSphere
import array_handler
import cv
import math

def cv2array(im):
  depth2dtype = {
        cv.IPL_DEPTH_8U: 'uint8',
        cv.IPL_DEPTH_8S: 'int8',
        cv.IPL_DEPTH_16U: 'uint16',
        cv.IPL_DEPTH_16S: 'int16',
        cv.IPL_DEPTH_32S: 'int32',
        cv.IPL_DEPTH_32F: 'float32',
        cv.IPL_DEPTH_64F: 'float64',
    }

  arrdtype=im.depth
  a = np.fromstring(
         im.tostring(),
         dtype=depth2dtype[im.depth],
         count=im.width*im.height*im.nChannels)
  a.shape = (im.height,im.width,im.nChannels)
  return a

def array2cv(a):
  dtype2depth = {
        'uint8':   cv.IPL_DEPTH_8U,
        'int8':    cv.IPL_DEPTH_8S,
        'uint16':  cv.IPL_DEPTH_16U,
        'int16':   cv.IPL_DEPTH_16S,
        'int32':   cv.IPL_DEPTH_32S,
        'float32': cv.IPL_DEPTH_32F,
        'float64': cv.IPL_DEPTH_64F,
    }
  try:
    nChannels = a.shape[2]
  except:
    nChannels = 1
  cv_im = cv.CreateImageHeader((a.shape[1],a.shape[0]),
          dtype2depth[str(a.dtype)],
          nChannels)
  cv.SetData(cv_im, a.tostring(),
             a.dtype.itemsize*nChannels*a.shape[1])
  return cv_im

class Visualizer:
	def visModelPly(self, plyFile):
		#----------------------------visualization--------------------------
		
		reader = vtk.vtkPLYReader()
		reader.SetFileName(plyFile)
		
		lineMapper2 = vtk.vtkPolyDataMapper()
		lineMapper2.SetInputConnection(reader.GetOutputPort())
		lineMapper2.Update()
		
		lineActor = vtk.vtkActor()
		lineActor.GetProperty().SetColor(0,1,0)
		lineActor.GetProperty().SetOpacity(1)
		lineActor.SetMapper(lineMapper2)

		# Create the usual graphics stuff.
		# Create the RenderWindow, Renderer and both Actors
		
		camera = vtk.vtkCamera()
		#camera.SetPosition(0,3,-6)
		camera.SetFocalPoint(0,0,0)
		
		ren = vtk.vtkRenderer()
		#ren.SetActiveCamera(camera)
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		ren.AddActor(lineActor)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
	
	def visModelsPly(self, plyFile, plyFile2):
		#----------------------------visualization--------------------------
		
		reader = vtk.vtkPLYReader()
		reader.SetFileName(plyFile)
		
		lineMapper2 = vtk.vtkPolyDataMapper()
		lineMapper2.SetInputConnection(reader.GetOutputPort())
		lineMapper2.Update()
		
		lineActor = vtk.vtkActor()
		lineActor.GetProperty().SetColor(0,1,0)
		lineActor.GetProperty().SetOpacity(1)
		lineActor.SetMapper(lineMapper2)
		
		reader2 = vtk.vtkPLYReader()
		reader2.SetFileName(plyFile2)
		
		lineMapper22 = vtk.vtkPolyDataMapper()
		lineMapper22.SetInputConnection(reader2.GetOutputPort())
		lineMapper22.Update()
		
		lineActor2 = vtk.vtkActor()
		lineActor2.GetProperty().SetColor(1,0,0)
		lineActor2.GetProperty().SetOpacity(1)
		lineActor2.SetMapper(lineMapper22)		

		# Create the usual graphics stuff.
		# Create the RenderWindow, Renderer and both Actors
		
		camera = vtk.vtkCamera()
		#camera.SetPosition(0,3,-6)
		camera.SetFocalPoint(0,0,0)
		
		ren = vtk.vtkRenderer()
		#ren.SetActiveCamera(camera)
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		ren.AddActor(lineActor)
		ren.AddActor(lineActor2)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
		
	def visModelAndContactPoints(self, plyFile, contactPoints, scaleFac, CoM, sphere, modelMapper=None, anotherMapper=None):
		#----------------------------visualization--------------------------
		
		#print len(contactPoints), contactPoints
		
		if modelMapper == None:
			reader = vtk.vtkPLYReader()
			reader.SetFileName(plyFile)
			
			trans = vtk.vtkTransform()
			trans.Translate(-CoM[0],-CoM[1],-CoM[2])
			tttf = vtk.vtkTransformFilter()
			tttf.SetTransform(trans)
			tttf.SetInputConnection(reader.GetOutputPort())
			
			lineMapper = vtk.vtkPolyDataMapper()
			lineMapper.SetInputConnection(tttf.GetOutputPort())
			lineMapper.Update()
			
			CoM = sphere.computeCenterOfMass(lineMapper.GetInput())
			#print CoM
			
			#scaleFac = 1
			#scaleFac /= 2.54
			#scaleFac = 25.4
			trans = vtk.vtkTransform()
			trans.Scale(scaleFac,scaleFac,scaleFac)
			
			tf = vtk.vtkTransformFilter()
			#tf.SetTransform(trans)
			tf.SetInputConnection(reader.GetOutputPort())
			
			lineMapper2 = vtk.vtkPolyDataMapper()
			lineMapper2.SetInputConnection(reader.GetOutputPort())
			lineMapper2.Update()
		else:
			lineMapper2 = modelMapper
		
		lineActor = vtk.vtkActor()
		lineActor.GetProperty().SetColor(0,1,0)
		lineActor.GetProperty().SetOpacity(0.5)
		lineActor.SetMapper(lineMapper2)

		coneSource = vtk.vtkConeSource()
		coneSource.SetResolution(6)
		
		mypoints = vtk.vtkPoints()
		poly = vtk.vtkPolyData()
		conn = vtk.vtkCellArray()
				
		for pp in contactPoints:
			mypoints.InsertNextPoint([pp[0],pp[1],pp[2]])
			
		for i in range(len(contactPoints)):
			a = conn.InsertNextCell(1)
			
		poly.SetPoints(mypoints)
		poly.SetVerts(conn)
		
		glyph = vtk.vtkGlyph3D()
		glyph.SetInput(poly)
		glyph.SetSourceConnection(coneSource.GetOutputPort())
		glyph.SetScaleFactor(0.1)

		spike = vtk.vtkPolyDataMapper()
		spike.SetInputConnection(glyph.GetOutputPort())
		
		spikeActor = vtk.vtkActor()
		spikeActor.SetMapper(spike)
		spikeActor.GetProperty().SetColor(1,0,0)
		
		# Create the usual graphics stuff.
		# Create the RenderWindow, Renderer and both Actors
		
		camera = vtk.vtkCamera()
		#camera.SetPosition(0,3,-6)
		camera.SetFocalPoint(0,0,0)
		
		ren = vtk.vtkRenderer()
		#ren.SetActiveCamera(camera)
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		ren.AddActor(lineActor)
		ren.AddActor(spikeActor)
				
		if anotherMapper != None:
			anotherLineActor = vtk.vtkActor()
			anotherLineActor.GetProperty().SetColor(0,0,1)
			anotherLineActor.GetProperty().SetOpacity(0.5)
			anotherLineActor.SetMapper(anotherMapper)
			ren.AddActor(anotherLineActor)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
		
	def visModel(self, mapper1, normal=None):
		act1 = vtk.vtkActor()
		act1.GetProperty().SetColor(0,1,0)
		act1.GetProperty().SetOpacity(1)
		act1.SetMapper(mapper1)
		
		ground = vtk.vtkPlaneSource()
		
		if normal != None:
			ground.SetNormal(normal)
		else:
			ground.SetNormal([0, 1, 0])
		
		ground.SetCenter([0, 0, 0])

		planeMapper2 = vtk.vtkPolyDataMapper()
		planeMapper2.SetInputConnection(ground.GetOutputPort())
		planeActor2 = vtk.vtkActor()
		planeActor2.SetMapper(planeMapper2)
		planeActor2.GetProperty().SetColor(1,0,0)
		planeActor2.GetProperty().SetOpacity(0.8)
		
		cam = vtk.vtkCamera()
		ren = vtk.vtkRenderer()
		ren.SetActiveCamera(cam)
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		axes = vtk.vtkAxes()
		axes.SetOrigin(0, 0, 0)
		axes.SetScaleFactor(0.1)

		axes_colors = vtk.vtkFloatArray()
		axes_colors.InsertNextValue(0.0)
		axes_colors.InsertNextValue(0.0)
		axes_colors.InsertNextValue(0.5)
		axes_colors.InsertNextValue(0.5)
		axes_colors.InsertNextValue(1.0)
		axes_colors.InsertNextValue(1.0)

		axes_data = axes.GetOutput()
		axes_data.Update()
		axes_data.GetPointData().SetScalars(axes_colors)

		axes_tubes = vtk.vtkTubeFilter()
		axes_tubes.SetInput (axes_data)
		axes_tubes.SetRadius (axes.GetScaleFactor () / 50.0)
		axes_tubes.SetNumberOfSides (6)

		axesMapper = vtk.vtkPolyDataMapper()
		axesMapper.SetScalarModeToUsePointData ()
		axesMapper.SetInput(axes_tubes.GetOutput())
		axesActor = vtk.vtkActor()
		axesActor.SetMapper(axesMapper)

		ren.AddActor(axesActor)

		ren.AddActor(act1)
		ren.AddActor(planeActor2)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()

	def visModelCam(self, mapper1, camera, color=(0,1,0), points=[], displayModel=True):

		act1 = vtk.vtkActor()
		act1.GetProperty().SetColor(color)
		act1.GetProperty().SetOpacity(1)
		act1.SetMapper(mapper1)
		
		ren = vtk.vtkRenderer()
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		if displayModel:
			ren.AddActor(act1)

		if len(points) > 0:
			pointsActor1 = vtk.vtkActor()
			pointsMapper1 = self.getPointsActor3D(points,0.005)
			pointsActor1.SetMapper(pointsMapper1)
			pointsActor1.GetProperty().SetColor(1,0,0)
			pointsActor1.GetProperty().SetOpacity(0.5)

			ren.AddActor(pointsActor1)

		axes = vtk.vtkAxes()
		axes.SetOrigin(0, 0, 0)
		axes.SetScaleFactor(0.1)

		axes_colors = vtk.vtkFloatArray()
		axes_colors.InsertNextValue(0.0)
		axes_colors.InsertNextValue(0.0)
		axes_colors.InsertNextValue(0.5)
		axes_colors.InsertNextValue(0.5)
		axes_colors.InsertNextValue(1.0)
		axes_colors.InsertNextValue(1.0)

		axes_data = axes.GetOutput()
		axes_data.Update()
		axes_data.GetPointData().SetScalars(axes_colors)

		axes_tubes = vtk.vtkTubeFilter()
		axes_tubes.SetInput (axes_data)
		axes_tubes.SetRadius (axes.GetScaleFactor () / 50.0)
		axes_tubes.SetNumberOfSides (6)

		axesMapper = vtk.vtkPolyDataMapper()
		axesMapper.SetScalarModeToUsePointData ()
		axesMapper.SetInput(axes_tubes.GetOutput())
		axesActor = vtk.vtkActor()
		axesActor.SetMapper(axesMapper)

		ren.AddActor(axesActor)
		ren.SetActiveCamera(camera)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
	
	def getPointsActor(self,points):
		coneSource = vtk.vtkConeSource()
		coneSource.SetResolution(6)
		
		mypoints = vtk.vtkPoints()
		poly = vtk.vtkPolyData()
		conn = vtk.vtkCellArray()
				
		for pp in points:
			mypoints.InsertNextPoint([pp[0],0,pp[1]])
			
		for i in range(len(points)):
			a = conn.InsertNextCell(1)
			
		poly.SetPoints(mypoints)
		poly.SetVerts(conn)
		
		glyph = vtk.vtkGlyph3D()
		glyph.SetInput(poly)
		glyph.SetSourceConnection(coneSource.GetOutputPort())
		glyph.SetScaleFactor(0.025)

		spike = vtk.vtkPolyDataMapper()
		spike.SetInputConnection(glyph.GetOutputPort())
		
		return spike
		
	def getPointsActor3D(self,points,scale=0.05):
		coneSource = vtk.vtkConeSource()
		coneSource.SetResolution(6)
		
		mypoints = vtk.vtkPoints()
		poly = vtk.vtkPolyData()
		conn = vtk.vtkCellArray()
		
		for pp in points:
			#print pp
			mypoints.InsertNextPoint([pp[0],pp[1],pp[2]])
			
		for i in range(len(points)):
			a = conn.InsertNextCell(1)
			
		poly.SetPoints(mypoints)
		poly.SetVerts(conn)
		
		glyph = vtk.vtkGlyph3D()
		glyph.SetInput(poly)
		glyph.SetSourceConnection(coneSource.GetOutputPort())
		glyph.SetScaleFactor(scale)

		spike = vtk.vtkPolyDataMapper()
		spike.SetInputConnection(glyph.GetOutputPort())
		
		return spike
		
	def getPlaneActor(self, normal):
		ground = vtk.vtkPlaneSource()
		ground.SetNormal(normal)
		ground.SetCenter([0, 0, 0])

		planeMapper2 = vtk.vtkPolyDataMapper()
		planeMapper2.SetInputConnection(ground.GetOutputPort())
		planeActor2 = vtk.vtkActor()
		planeActor2.SetMapper(planeMapper2)
		return planeActor2

	def displayMapperAndPoints(self, points, mapper):
		pointsActor1 = vtk.vtkActor()
		pointsMapper1 = self.getPointsActor3D(points)
		pointsActor1.SetMapper(pointsMapper1)
		pointsActor1.GetProperty().SetColor(1,0,0)
		pointsActor1.GetProperty().SetOpacity(0.5)
		
		ren = vtk.vtkRenderer()
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		act1 = vtk.vtkActor()
		act1.GetProperty().SetColor(1,0,0)
		act1.GetProperty().SetOpacity(0.5)
		act1.SetMapper(mapper)
		
		ren.AddActor(pointsActor1)
		ren.AddActor(act1)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()

	def VTKtoOpenCVImg(self, im):
		numPyArray = array_handler.vtk2array(im)
		size = int(math.sqrt(numPyArray.shape[0]))
		a = np.reshape(numPyArray, (size, size, 3))
		
		im = cv.CreateImage( (size, size) , 8, 3)
		im = array2cv(a)
		cv.Flip(im,im, 0)
		
		#finally convert to binary image!
		binary = cv.CreateImage( cv.GetSize(im), cv.IPL_DEPTH_8U, 1)
		cv.CvtColor(im , binary, cv.CV_RGB2GRAY)
		return binary
		
	def getViewUp(self, normal):
		if normal[1] == 1:
			return [1,0,0]
		else:
			return [0,1,0]
			
	
	def getImageProjectionFromMapperSettingNormal(self, mapper, normal, fileName, CoM, stereo=False):
		trans2scale = vtk.vtkTransform()
		trans2scale.Scale(1,1,1)
		transFilter22 = vtk.vtkTransformFilter()
		transFilter22.SetTransform(trans2scale)
		transFilter22.SetInput(mapper.GetInput())
		mapper2Scaled = vtk.vtkPolyDataMapper()
		mapper2Scaled.SetInputConnection(transFilter22.GetOutputPort())
		mapper2Scaled.Update()
		
		actor = vtk.vtkActor()
		actor.GetProperty().SetColor(0,0,0)
		actor.SetMapper(mapper2Scaled)
		
		camera = vtk.vtkCamera()
		camera.SetFocalPoint(CoM)
		camera.SetPosition(np.array(normal) * 3)
		camera.SetParallelProjection(1)
		camera.SetViewUp(self.getViewUp(normal))
		
		ren = vtk.vtkRenderer()
		ren.SetActiveCamera(camera)
		ren.AddActor(actor)
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(128,128)		
		ren.Render()
		
		winToImageFilter = vtk.vtkWindowToImageFilter()
		winToImageFilter.SetInput(renWin);

		renWin.Finalize()
		winToImageFilter.Update()

		#transform vtk image to opencv image
		binary = self.VTKtoOpenCVImg(winToImageFilter.GetOutput().GetPointData().GetScalars())
		
		if stereo:
			#dilate image to make it cleaner
			cv.Not(binary,binary)
			cv.Threshold(binary, binary, 20, 255, cv.CV_THRESH_BINARY)
			cv.Dilate(binary,binary,None,2)
			#cv.Dilate(binary,binary,None,1)
			cv.Erode(binary,binary,None,1)
			cv.Erode(binary,binary,None,1)
			cv.Not(binary,binary)
		
		cv.SaveImage(fileName + ".jpg", binary)
		#jw = vtk.vtkJPEGWriter();
		#jw.SetInput(winToImageFilter.GetOutput());
		
		#jw.SetFileName(fileName + ".jpg");
		#jw.Write();
		
		return binary

	#kept in older versions!!	
	#def displayProjectWithMainAxes(self, points1, points2, axes1, axes2, mapper1, mapper2):
	#def displayProjectedWithMainAxes(self, points1, axes1, mapper1):
		
	def displayProjectedWithMapper(self, points1, mapper1):
		pointsActor1 = vtk.vtkActor()
		pointsMapper1 = self.getPointsActor(points1)
		pointsActor1.SetMapper(pointsMapper1)
		pointsActor1.GetProperty().SetColor(1,0,0)
		pointsActor1.GetProperty().SetOpacity(0.5)
		ren = vtk.vtkRenderer()
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)
		
		act1 = vtk.vtkActor()
		act1.GetProperty().SetColor(1,1,0)
		act1.GetProperty().SetOpacity(0.5)
		act1.SetMapper(mapper1)
		
		ren.AddActor(pointsActor1)
		#ren.AddActor(planeActor1)
		#ren.AddActor(planeActor11)
		ren.AddActor(act1)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
	
	def getPolygonMapper(self, points1, height):
		polygonPoints = vtk.vtkPoints()
		polygonPoints.SetNumberOfPoints(len(points1))
		aPolygon = vtk.vtkPolygon()
		aPolygon.GetPointIds().SetNumberOfIds(len(points1))
		i=0
		for p1 in points1:
			polygonPoints.InsertPoint(i, p1[0], height, p1[1])
			aPolygon.GetPointIds().SetId(i, i)
			i+=1
			
		aPolygonGrid = vtk.vtkUnstructuredGrid()
		aPolygonGrid.Allocate(1, 1)
		aPolygonGrid.InsertNextCell(aPolygon.GetCellType(), aPolygon.GetPointIds())
		aPolygonGrid.SetPoints(polygonPoints)
		aPolygonMapper = vtk.vtkDataSetMapper()
		aPolygonMapper.SetInput(aPolygonGrid)
		return aPolygonMapper		
	
	def visualizePolygon(self, points1):
		
		aPolygonMapper = self.getPolygonMapper(points1, 0.1)
		
		aPolygonActor = vtk.vtkActor()
		aPolygonActor.SetMapper(aPolygonMapper)
		aPolygonActor.AddPosition(0, 0, 0)
		aPolygonActor.GetProperty().SetDiffuseColor(0, 0, 0)
				
		ren = vtk.vtkRenderer()
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		ren.AddActor(aPolygonActor)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
	
	def getMapperFromPointCloud(self, points):
		pointsMapper = self.getPointsActor3D(points, 0.005)
		return pointsMapper
		
	def visStereoAnd3DModel(self, points, plyFile):
		pointsActor1 = vtk.vtkActor()
		pointsMapper1 = self.getPointsActor3D(points, 0.005)
		pointsActor1.SetMapper(pointsMapper1)
		pointsActor1.GetProperty().SetColor(1,0,0)
		pointsActor1.GetProperty().SetOpacity(0.5)
		
		reader = vtk.vtkPLYReader()
		reader.SetFileName(plyFile)
		
		lineMapper2 = vtk.vtkPolyDataMapper()
		lineMapper2.SetInputConnection(reader.GetOutputPort())
		lineMapper2.Update()
		
		lineActor = vtk.vtkActor()
		lineActor.GetProperty().SetColor(0,1,0)
		lineActor.GetProperty().SetOpacity(1)
		lineActor.SetMapper(lineMapper2)
		
		ren = vtk.vtkRenderer()
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)
		
		ren.AddActor(lineActor)
		ren.AddActor(pointsActor1)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
	
	def visModels(self, mapper1, mapper2):
		
		bounds = mapper1.GetInput().GetBounds() #compute bounds again
		#print bounds
		
		act1 = vtk.vtkActor()
		act1.GetProperty().SetColor(0,1,0)
		act1.GetProperty().SetOpacity(1)
		act1.SetMapper(mapper1)
		
		act2 = vtk.vtkActor()
		act2.GetProperty().SetColor(1,0,0)
		act2.GetProperty().SetOpacity(0.5)
		act2.SetMapper(mapper2)
		
		ground = vtk.vtkPlaneSource()
		ground.SetNormal([0, 1, 0])
		ground.SetCenter([0, 0, 0])

		planeMapper2 = vtk.vtkPolyDataMapper()
		planeMapper2.SetInputConnection(ground.GetOutputPort())
		planeActor2 = vtk.vtkActor()
		planeActor2.SetMapper(planeMapper2)
		planeActor2.GetProperty().SetColor(1,0,1)
		planeActor2.GetProperty().SetOpacity(0.5)
		planeActor2.SetScale(max(bounds)*3)
		
		ren = vtk.vtkRenderer()
		camera = vtk.vtkCamera()
		camera.SetFocalPoint([0,0,0])
		camera.SetPosition([0,4,0])
		camera.SetParallelProjection(1)
		#camera.SetViewUp([1,0,0])
		#ren.SetActiveCamera(camera)
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		#act1.RotateX(180)
		#act2.RotateX(180)

		ren.AddActor(act1)
		ren.AddActor(act2)
		ren.AddActor(planeActor2)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
		
	def visModelsAndPoints(self, mapper1, mapper2, points):
		
		bounds = mapper1.GetInput().GetBounds() #compute bounds again
		#print bounds
		
		act1 = vtk.vtkActor()
		act1.GetProperty().SetColor(0,1,0)
		act1.GetProperty().SetOpacity(1)
		act1.SetMapper(mapper1)
		
		act2 = vtk.vtkActor()
		act2.GetProperty().SetColor(1,0,0)
		act2.GetProperty().SetOpacity(0.5)
		act2.SetMapper(mapper2)
		
		ground = vtk.vtkPlaneSource()
		ground.SetNormal([0, 1, 0])
		ground.SetCenter([0, 0, 0])

		planeMapper2 = vtk.vtkPolyDataMapper()
		planeMapper2.SetInputConnection(ground.GetOutputPort())
		planeActor2 = vtk.vtkActor()
		planeActor2.SetMapper(planeMapper2)
		planeActor2.GetProperty().SetColor(1,0,1)
		planeActor2.GetProperty().SetOpacity(0.3)
		planeActor2.SetScale(max(bounds)*3)
		
		ren = vtk.vtkRenderer()
		camera = vtk.vtkCamera()
		camera.SetFocalPoint([0,0,0])
		camera.SetPosition([0,4,0])
		camera.SetParallelProjection(1)
		#camera.SetViewUp([1,0,0])
		#ren.SetActiveCamera(camera)
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		pointActor = vtk.vtkActor()
		pointsMapper1 = self.getPointsActor3D(points, 0.05)
		pointActor.SetMapper(pointsMapper1)
		pointActor.GetProperty().SetColor(1,0,0)
		pointActor.GetProperty().SetOpacity(0.5)

		ren.AddActor(act1)
		ren.AddActor(act2)
		ren.AddActor(planeActor2)
		ren.AddActor(pointActor)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
		
	def visModelsAndPointsAndDescriptor(self, mapper1, mapper2, points, descActor1,descActor2):
		
		bounds = mapper1.GetInput().GetBounds() #compute bounds again
		#print bounds
		
		act1 = vtk.vtkActor()
		act1.GetProperty().SetColor(0,1,0)
		act1.GetProperty().SetOpacity(0.5)
		act1.SetMapper(mapper1)
		
		act2 = vtk.vtkActor()
		act2.GetProperty().SetColor(1,0,0)
		act2.GetProperty().SetOpacity(0.5)
		act2.SetMapper(mapper2)
		
		ground = vtk.vtkPlaneSource()
		ground.SetNormal([0, 1, 0])
		ground.SetCenter([0, 0, 0])

		planeMapper2 = vtk.vtkPolyDataMapper()
		planeMapper2.SetInputConnection(ground.GetOutputPort())
		planeActor2 = vtk.vtkActor()
		planeActor2.SetMapper(planeMapper2)
		planeActor2.GetProperty().SetColor(1,0,1)
		planeActor2.GetProperty().SetOpacity(0.3)
		planeActor2.SetScale(max(bounds)*3)
		
		ren = vtk.vtkRenderer()
		camera = vtk.vtkCamera()
		camera.SetFocalPoint([0,0,0])
		camera.SetPosition([0,4,0])
		camera.SetParallelProjection(1)
		#camera.SetViewUp([1,0,0])
		#ren.SetActiveCamera(camera)
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		pointActor = vtk.vtkActor()
		pointsMapper1 = self.getPointsActor3D(points, 0.05)
		pointActor.SetMapper(pointsMapper1)
		pointActor.GetProperty().SetColor(1,0,0)
		pointActor.GetProperty().SetOpacity(0.5)

		#print descActor1.GetPosition()
		#print act2.GetPosition()
		descActor1.SetPosition(-0.75,0.25,0.75)
		act2.SetPosition(-0.75,0.25,0.75)
		
		ren.AddActor(act1)
		ren.AddActor(act2)
		#ren.AddActor(planeActor2)
		#ren.AddActor(pointActor)
		ren.AddActor(descActor1)
		ren.AddActor(descActor2)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
		
	def visModelsWithOutPlane(self, mapper1, mapper2):
		
		bounds = mapper1.GetInput().GetBounds() #compute bounds again
		#print bounds
		
		act1 = vtk.vtkActor()
		act1.GetProperty().SetColor(0,1,0)
		act1.GetProperty().SetOpacity(1)
		act1.SetMapper(mapper1)
		
		act2 = vtk.vtkActor()
		act2.GetProperty().SetColor(1,0,0)
		act2.GetProperty().SetOpacity(1)
		act2.SetMapper(mapper2)
		
		ren = vtk.vtkRenderer()
		#ren.SetActiveCamera(camera)
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		ren.AddActor(act1)
		ren.AddActor(act2)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
		
	def visModelsAndTwoPlanes(self, mapper1, mapper2, n1,n2):
		
		bounds = mapper1.GetInput().GetBounds() #compute bounds again
		#print bounds
		
		act1 = vtk.vtkActor()
		act1.GetProperty().SetColor(0,1,0)
		act1.GetProperty().SetOpacity(0.5)
		act1.SetMapper(mapper1)
		
		act2 = vtk.vtkActor()
		act2.GetProperty().SetColor(1,0,0)
		act2.GetProperty().SetOpacity(0.5)
		act2.SetMapper(mapper2)
		
		ground = vtk.vtkPlaneSource()
		ground.SetNormal([0, 1, 0])
		ground.SetCenter([0, 0, 0])

		planeMapper2 = vtk.vtkPolyDataMapper()
		planeMapper2.SetInputConnection(ground.GetOutputPort())
		planeActor2 = vtk.vtkActor()
		planeActor2.SetMapper(planeMapper2)
		planeActor2.GetProperty().SetColor(1,0,1)
		planeActor2.GetProperty().SetOpacity(0.5)
		planeActor2.SetScale(max(bounds)*3)
		
		ren = vtk.vtkRenderer()
		renWin = vtk.vtkRenderWindow()
		renWin.AddRenderer(ren)
		iren = vtk.vtkRenderWindowInteractor()
		iren.SetRenderWindow(renWin)

		act1.RotateX(180)
		act2.RotateX(180)

		ren.AddActor(act1)
		ren.AddActor(act2)
		#ren.AddActor(planeActor2)
		
		ground1 = vtk.vtkPlaneSource()
		ground1.SetNormal(n1)
		ground1.SetCenter([0, 0, 0])

		planeGround1 = vtk.vtkPolyDataMapper()
		planeGround1.SetInputConnection(ground1.GetOutputPort())
		planeActorG1 = vtk.vtkActor()
		planeActorG1.SetMapper(planeGround1)
		planeActorG1.GetProperty().SetColor(0,1,0)
		planeActorG1.GetProperty().SetOpacity(0.5)
		
		ground2 = vtk.vtkPlaneSource()
		ground2.SetNormal(n2)
		ground2.SetCenter([0, 0, 0])

		planeGround2 = vtk.vtkPolyDataMapper()
		planeGround2.SetInputConnection(ground2.GetOutputPort())
		planeActorG2 = vtk.vtkActor()
		planeActorG2.SetMapper(planeGround2)
		planeActorG2.GetProperty().SetColor(1,0,0)
		planeActorG2.GetProperty().SetOpacity(0.5)
		
		ren.AddActor(planeActorG1)
		ren.AddActor(planeActorG2)
		
		ren.SetBackground(1, 1, 1)
		renWin.SetSize(600, 600)
		
		iren.Initialize()
		renWin.Render()
		iren.Start()

		renWin.Finalize()
