#!/usr/bin/env python

import os
import time
import vtk
import sys 
import time
import numpy
import random
from math import sqrt
import pickle
import threading
import os
import math
import copy

sys.path.append("..")
from utils.tesselatedSphere import TesselatedSphere 
from utils.visualize import Visualizer

import ground_plane.utils as utils

def writeMatrix(matrix,fileD):
  for i in xrange(4):
    for j in xrange(4):
      fileD.write(str(matrix.GetElement(i,j)) + " ")
    fileD.write("\n")

def copyMatrix(matrix):
  copy = vtk.vtkMatrix4x4()
  for i in xrange(4):
    for j in xrange(4):
      copy.SetElement(i,j,matrix.GetElement(i,j))

  return copy

def vtkToNumpy(vtk_matrix):
  np_matrix = numpy.zeros((4,4))
  for i in xrange(4):
    for j in xrange(4):
      np_matrix[i,j] = float(vtk_matrix.GetElement(i,j))
  
  return np_matrix

def sample_views_tesselated_sphere(source_file, outfile):
  #if not os.path.isdir(view_directory):
  #  os.mkdir(view_directory)
  
  sphereLevel = 0
  sphere = TesselatedSphere(sphereLevel, 0)
  sphere.construct()
  vis = Visualizer()
  reader1 = vtk.vtkPLYReader()
  reader1.SetFileName(source_file)

  ren = vtk.vtkRenderer()
  renWin = vtk.vtkRenderWindow()
  renWin.AddRenderer(ren)
  iren = vtk.vtkRenderWindowInteractor()
  ren.SetBackground(1.0, 0.0, 0.0)
  xres = 350
  yres = 350
  renWin.SetSize(xres, yres)
  renWin.SetPosition(1000,200)
  worldPicker = vtk.vtkWorldPointPicker ()
  resolution =1
  viewAngle = 45
  ren.ResetCamera()
  cam = ren.GetActiveCamera()
  cam.SetFocalPoint(0,0,0)
  cam.SetClippingRange(0,10)
  cam.SetViewAngle(viewAngle)

  stlMapper = vtk.vtkPolyDataMapper()
  stlMapper.SetInputConnection(reader1.GetOutputPort())
  stlMapper.Update()

  CoM = sphere.computeCenterOfMass(stlMapper.GetInput())
  transCenter = vtk.vtkTransform()
  transCenter.Translate(-CoM[0],-CoM[1],-CoM[2])
  matrixCenter = transCenter.GetMatrix()
  transFilterCenter = vtk.vtkTransformFilter()
  transFilterCenter.SetTransform(transCenter)
  transFilterCenter.SetInputConnection(reader1.GetOutputPort())

  stlMapper = vtk.vtkPolyDataMapper()
  stlMapper.SetInputConnection(transFilterCenter.GetOutputPort())
  stlMapper.Update()

  bb = stlMapper.GetBounds()
  ms = max( (abs(bb[0]-bb[1]),abs(bb[2]-bb[3]),abs(bb[4]-bb[5]) ) )
  max_side=0.5
  scale_factor = max_side/ms
  vscale = vtk.vtkTransform()

  vscale.Scale(scale_factor,scale_factor,scale_factor)
  matrixScale = vscale.GetMatrix()

  transFilterScale = vtk.vtkTransformFilter()
  transFilterScale.SetTransform(vscale)
  transFilterScale.SetInputConnection(transFilterCenter.GetOutputPort())

  stlMapper = vtk.vtkPolyDataMapper()
  stlMapper.SetInputConnection(transFilterScale.GetOutputPort())
  stlMapper.Update()

  #print len(sphere.polygons)
  radius=1
  firstPos = sphere.polygons[sphereLevel][0].getCenter()*radius
  gen_views = []
  views_transforms = []
  view = 0

  #loop over the tesselated sphere to get camera positions
  for polygon in sphere.polygons[sphereLevel]:
    #print polygon.getCenter()
    cam.SetPosition(polygon.getCenter()*radius)

    pts = vtk.vtkPoints()
    ren.SetActiveCamera(cam)
    ren.ResetCameraClippingRange()
    stlActorTrans = vtk.vtkActor()
    stlActorTrans.SetMapper(stlMapper)
    ren.AddActor(stlActorTrans)
    ren.SetBackground(1.0, 0.0, 0.0)

    renWin.Render()

    w2i = vtk.vtkWindowToImageFilter()
    w2i.SetInput(renWin)
    imD  = w2i.GetOutput()  
    imD.Update()
    imD.SetNumberOfScalarComponents(3)

    worldPicker.Pick(5,5,0,ren)
    p1 = numpy.array(worldPicker.GetPickPosition())
    worldPicker.Pick(xres-5,5,0,ren)
    p2 = numpy.array(worldPicker.GetPickPosition())
    worldPicker.Pick(5,yres-5,0,ren)
    p3 = numpy.array(worldPicker.GetPickPosition())

    pn = numpy.cross(p2-p1, p3-p1 )
    pn = pn / sqrt(numpy.dot(pn,pn.conj()))
    d = numpy.dot(pn,p1.conj())

    narr = []
    
    for x in range(0,xres,resolution):
      for y in range(0,yres,resolution):
        if imD.GetScalarComponentAsFloat(x,y,0,0) == 255 and imD.GetScalarComponentAsFloat(x,y,0,1)  == 0 and imD.GetScalarComponentAsFloat(x,y,0,2)  == 0 :
          continue
        else:
          worldPicker.Pick(x,y,0,ren)
          nx,ny,nz = worldPicker.GetPickPosition()
          #THIS is causing the missing lines, if you comment out, ... yea, see the odd behaviour of VTK ;)
          if abs(numpy.dot([nx,ny,nz],pn.conj()) - d ) < 0.001:
            continue

          narr.append([nx,ny,nz])

    #vis.visModelCam(stlMapper, cam, (1,0,1), narr, True)

    gen_views.append(narr) #views in camera coordinates
    view += 1
  #print views_transforms[0], views_transforms[1]

  whole_mesh = list()
  for i in xrange(len(gen_views)):
    for j in xrange(len(gen_views[i])):
      whole_mesh.append(gen_views[i][j])

  with open (outfile, 'w') as f:
    f.write ("COLUMNS x y z\n")
    f.write ("POINTS %i\n" % len(whole_mesh))
    f.write ("DATA ascii\n")
    for point in whole_mesh:
      for dim in point:
        f.write ("%f " % dim)
      f.write ("\n")
      

  #print len(whole_mesh)    
  vis.visModelCam(stlMapper, cam, (1,0,1), whole_mesh, True)

if __name__ == "__main__":
  if (len(sys.argv) == 3):
    sample_views_tesselated_sphere(sys.argv[1], sys.argv[2])
  else:
    print "usage: %s infile outfile" % sys.argv[0]


