#! /usr/bin/env python

import os
import time
import vtk
import sys 
import time
import numpy
import random
from math import sqrt
import pickle
import threading
import os
import math
import copy

sys.path.append("..")
from utils.tesselatedSphere import TesselatedSphere 

def writeMatrix(matrix,fileD):
	for i in xrange(4):
		for j in xrange(4):
			fileD.write(str(matrix.GetElement(i,j)) + " ")
		fileD.write("\n")

def copyMatrix(matrix):
	copy = vtk.vtkMatrix4x4()
	for i in xrange(4):
		for j in xrange(4):
			copy.SetElement(i,j,matrix.GetElement(i,j))

	return copy

def vtkToNumpy(vtk_matrix):
	np_matrix = numpy.zeros((4,4))
	for i in xrange(4):
		for j in xrange(4):
			np_matrix[i,j] = float(vtk_matrix.GetElement(i,j))
	
	return np_matrix

def sample_views_tesselated_sphere(source_file):
	sphereLevel = 0
	sphere = TesselatedSphere(sphereLevel, 0)
	sphere.construct()
	reader1 = vtk.vtkPLYReader()
	reader1.SetFileName(source_file)

	ren = vtk.vtkRenderer()
	renWin = vtk.vtkRenderWindow()
	renWin.AddRenderer(ren)
	iren = vtk.vtkRenderWindowInteractor()
	ren.SetBackground(1.0, 0.0, 0.0)
	xres = 150
	yres = 150
	renWin.SetSize(xres, yres)
	renWin.SetPosition(1000,200)
	worldPicker = vtk.vtkWorldPointPicker ()
	resolution =1
	viewAngle = 45
	ren.ResetCamera()
	cam = ren.GetActiveCamera()
	cam.SetFocalPoint(0,0,0)
	cam.SetClippingRange(0,10)
	cam.SetViewAngle(viewAngle)

	stlMapper = vtk.vtkPolyDataMapper()
	stlMapper.SetInputConnection(reader1.GetOutputPort())
	stlMapper.Update()

	CoM = sphere.computeCenterOfMass(stlMapper.GetInput())
	transCenter = vtk.vtkTransform()
	transCenter.Translate(-CoM[0],-CoM[1],-CoM[2])
	matrixCenter = transCenter.GetMatrix()
	transFilterCenter = vtk.vtkTransformFilter()
	transFilterCenter.SetTransform(transCenter)
	transFilterCenter.SetInputConnection(reader1.GetOutputPort())

	stlMapper = vtk.vtkPolyDataMapper()
	stlMapper.SetInputConnection(transFilterCenter.GetOutputPort())
	stlMapper.Update()

	bb = stlMapper.GetBounds()
	ms = max( (abs(bb[0]-bb[1]),abs(bb[2]-bb[3]),abs(bb[4]-bb[5]) ) )
	max_side=0.5
	scale_factor = max_side/ms
	vscale = vtk.vtkTransform()

	vscale.Scale(scale_factor,scale_factor,scale_factor)
	matrixScale = vscale.GetMatrix()

	transFilterScale = vtk.vtkTransformFilter()
	transFilterScale.SetTransform(vscale)
	transFilterScale.SetInputConnection(transFilterCenter.GetOutputPort())

	stlMapper = vtk.vtkPolyDataMapper()
	stlMapper.SetInputConnection(transFilterScale.GetOutputPort())
	stlMapper.Update()

	print len(sphere.polygons)
	radius=1
	firstPos = sphere.polygons[sphereLevel][0].getCenter()*radius
	gen_views = []
	views_transforms = []
	view = 0

	#loop over the tesselated sphere to get camera positions
	for polygon in sphere.polygons[sphereLevel]:
		#print polygon.getCenter()
		cam.SetPosition(polygon.getCenter()*radius)

		camC = vtk.vtkCamera()
		camC.SetFocalPoint(0,0,0)
		camC.SetViewUp(0,1,0)
		camC.SetPosition(firstPos)
		camC.SetViewAngle(viewAngle)
		camC.Modified()

		transROT = vtk.vtkTransform()
		transROT.Identity()
		#transROT.RotateZ(view*30)
		transROT.Concatenate(camC.GetViewTransformObject().GetInverse().GetMatrix())
		transROT.Concatenate(cam.GetViewTransformObject().GetMatrix())
		print "determinant:", numpy.linalg.det(vtkToNumpy(transROT.GetMatrix()))

		transFilter11 = vtk.vtkTransformFilter()
		transFilter11.SetTransform(transROT)
		#transFilter11.SetInputConnection(transFilterCenter.GetOutputPort())
		transFilter11.SetInputConnection(transFilterScale.GetOutputPort())
		matrixRotModel = transROT.GetMatrix()

		transMapper = vtk.vtkPolyDataMapper()
		transMapper.SetInputConnection(transFilter11.GetOutputPort())
		transMapper.Update()

		#translate points so camera is at (0,0,0)
		translation = vtk.vtkTransform()
		translation.Translate(firstPos*-1)
		matrixTranslation = translation.GetMatrix()

		translationFilter1 = vtk.vtkTransformFilter()
		translationFilter1.SetTransform(translation)
		translationFilter1.SetInputConnection(transFilter11.GetOutputPort())

		transMapper = vtk.vtkPolyDataMapper()
		transMapper.SetInputConnection(translationFilter1.GetOutputPort())
		transMapper.Update()

		camC.SetPosition((0,0,0))
		camC.SetFocalPoint(firstPos*-1)
		#vis.visModelCam(stlMapper, cam)
		#vis.visModelCam(transMapper, camC, (1,0,0))

		transMapperTranslated = vtk.vtkPolyDataMapper()
		transMapperTranslated.SetInputConnection(translationFilter1.GetOutputPort())
		transMapperTranslated.Update()

		#camC is now located at (0,0,0) and gives the same view as cam
		viewT = camC.GetViewTransformMatrix()
		pts = vtk.vtkPoints()
		ren.SetActiveCamera(camC)
		ren.ResetCameraClippingRange()
		stlActorTrans = vtk.vtkActor()
		stlActorTrans.SetMapper(transMapperTranslated)
		ren.AddActor(stlActorTrans)
		ren.SetBackground(1.0, 0.0, 0.0)

		renWin.Render()

		w2i = vtk.vtkWindowToImageFilter()
		w2i.SetInput(renWin)
		imD  = w2i.GetOutput()	
		imD.Update()
		imD.SetNumberOfScalarComponents(3)

		writer = vtk.vtkPNGWriter()
		writer.SetInput(imD)
		writer.SetFileName("view_" + str(view) + ".png")
		writer.Write()

		#ren_cam = vtk.vtkRenderer()
		#ren_cam.SetActiveCamera(cam)
		#ren_cam.ResetCameraClippingRange()
		#actor_static = vtk.vtkActor()
		#actor_static.SetMapper(stlMapper)
		#ren_cam.AddActor(actor_static)
		#ren_cam.SetBackground(1.0, 0.0, 0.0)

		#renWinCam = vtk.vtkRenderWindow()
		#renWinCam.AddRenderer(ren_cam)
		#renWinCam.Render()

		#w2i_cam = vtk.vtkWindowToImageFilter()
		#w2i_cam.SetInput(renWinCam)
		#imD_cam  = w2i.GetOutput()	
		#imD_cam.Update()
		#imD_cam.SetNumberOfScalarComponents(3)

		
		#writer = vtk.vtkPNGWriter()
		#writer.SetInput(imD_cam)
		#writer.SetFileName("view_cam_" + str(view) + ".png")
		#writer.Write()
		
		worldPicker.Pick(5,5,0,ren)
		p1 = numpy.array(worldPicker.GetPickPosition())
		worldPicker.Pick(xres-5,5,0,ren)
		p2 = numpy.array(worldPicker.GetPickPosition())
		worldPicker.Pick(5,yres-5,0,ren)
		p3 = numpy.array(worldPicker.GetPickPosition())

		pn = numpy.cross(p2-p1, p3-p1 )
		pn = pn / sqrt(numpy.dot(pn,pn.conj()))
		d = numpy.dot(pn,p1.conj())

		narr = []

		matrixTranslationCopy = copyMatrix(matrixTranslation)
		backToRealScale = vtk.vtkTransform()
		backToRealScale.PostMultiply()
		backToRealScale.Identity()
		backToRealScale.Concatenate(matrixScale) #scale to real scale			
		backToRealScale.Concatenate(matrixTranslationCopy) #translate back to (0,0,0)		
		backToRealScale.Inverse()
		backToRealScale.Modified()
		backToRealScale.Concatenate(matrixTranslation)
		backToRealScale.Modified()
		
		for x in range(0,xres,resolution):
			for y in range(0,yres,resolution):
				if imD.GetScalarComponentAsFloat(x,y,0,0) == 255 and imD.GetScalarComponentAsFloat(x,y,0,1)  == 0 and imD.GetScalarComponentAsFloat(x,y,0,2)  == 0 :
					continue
				else:
					worldPicker.Pick(x,y,0,ren)
					nx,ny,nz = worldPicker.GetPickPosition()
					#THIS is causing the missing lines, if you comment out, ... yea, see the odd behaviour of VTK ;)
					if abs(numpy.dot([nx,ny,nz],pn.conj()) - d ) < 0.001:
						continue

					#we want the coordinates in camera coordinates! z forward, y down and x to the right
					#points are in world coordinates
					pWorld = [nx,ny,nz,1]

					pWorldTrans = backToRealScale.GetMatrix().MultiplyPoint(pWorld)
					pCamera = numpy.array(viewT.MultiplyPoint(pWorldTrans))
					narr.append([pCamera[0],-pCamera[1],-pCamera[2]]) #rotation about the x axes (180 degrees) to match view up + flip x coordinate due to VIEW VTK coordinate system...

		kinect = vtk.vtkCamera()
		kinect.SetFocalPoint(0,0,5)
		kinect.SetViewUp(0,-1,0)
		kinect.SetPosition((0,0,0))
		kinect.Modified()

		stdCam = vtk.vtkCamera()
		ren.RemoveActor(stlActorTrans)

		#create the whole transformation, from object coordinates to camera coordinates
		transOCtoCC = vtk.vtkTransform()
		transOCtoCC.PostMultiply()
		transOCtoCC.Identity()
		#print matrixCenter
		transOCtoCC.Concatenate(matrixCenter)
		transOCtoCC.Concatenate(matrixRotModel)
		transOCtoCC.Concatenate(matrixTranslation)
		transOCtoCC.Concatenate(viewT)
	
		#kinect transform
		kinectTMatrix = vtk.vtkMatrix4x4()
		kinectTMatrix.Identity()
		kinectTMatrix.SetElement(0,0,1)
		kinectTMatrix.SetElement(1,1,-1)
		kinectTMatrix.SetElement(2,2,-1)

		transOCtoCC.Concatenate(kinectTMatrix)
		transOCtoCC.Modified()

		print "determinant:", numpy.linalg.det(vtkToNumpy(transOCtoCC.GetMatrix()))
		reader1 = vtk.vtkPLYReader()
		reader1.SetFileName(source_file)

		finalFilter = vtk.vtkTransformFilter()
		#transOCtoCC.Inverse()
		finalFilter.SetTransform(transOCtoCC)
		finalFilter.SetInputConnection(reader1.GetOutputPort())

		mapper1 = vtk.vtkPolyDataMapper()
		mapper1.SetInputConnection(finalFilter.GetOutputPort())
		mapper1.Update()
		#vis.visModel(mapper1)

		#vis.visModelCam(mapper1, kinect, (1,0,1), narr, True)

		#check if view already exists by comparing the rendered image!

		gen_views.append(narr) #views in camera coordinates
		#print transOCtoCC.GetMatrix()
		views_transforms.append(transOCtoCC.GetMatrix()) #transformation

		view += 1
	#print views_transforms[0], views_transforms[1]

	#quit()
	return [gen_views,views_transforms]
