#
# Software License Agreement (BSD License)
#
#  Copyright (c) 2010, Willow Garage, Inc.
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of Willow Garage, Inc. nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
#  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
#  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
#  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.
#
# \author Aitor Aldoma
#
# Connects to the household_objects and call for
# generate_views_vfh for the models for which views and vfh descriptors
# havent been yet generated...
# To execute this script: python scripts/household_objects_db_update.py 
#

import psycopg2
import sys
import os
import pickle
import dbutils

try:
	connection = psycopg2.connect(database="household_objects", user="willow", password="willow", host="wgs36")
except:
	print "I am unable to connect to the database."

cur = connection.cursor()

DB_PATH=dbutils.getPlyModelsRoot(cur)
ITERATION = int(sys.argv[2])

#read models in a directory
directory = sys.argv[1]
query = ""
for f in os.listdir(directory):
	if os.path.isfile(os.path.join(directory, f)) and f.endswith(".ply"):
		db_path = "ply/" + f
		query = query + "fp.file_path_path = '" + db_path + "' OR "

print query[0:-4]
final_query = "SELECT fp.file_path_path,om.original_model_id from original_model as om, file_path as fp WHERE fp.original_model_id = om.original_model_id AND fp.file_type = 'GEOMETRY_BINARY_PLY' AND (" + query[0:-4] + ") ORDER BY om.original_model_id "

try:
	cur.execute(final_query)
	#9463 - all except the new models..
	#cur.execute("""SELECT fp.file_path_path,om.original_model_id from original_model as om, file_path as fp WHERE fp.original_model_id = om.original_model_id AND fp.file_type = 'GEOMETRY_BINARY_PLY' ORDER BY om.original_model_id""") 
except:
	print "I can't SELECT from original_model"

rows = cur.fetchall()
for row in rows:
	fnstr = DB_PATH + row[0]
	model_id = row[1] 

	cur.execute("""SELECT scaled_model_id,scaled_model_scale from scaled_model WHERE original_model_id = """ + str(model_id))
	scaled_models = cur.fetchall()
	for scaled_model in scaled_models:
		#parameters are the path to the ply file, the scaled_model_id and the scale_factor to apply to the ply model
		if not dbutils.checkViewsGenerated(cur, scaled_model[0], ITERATION):
			scaled_factor = float(scaled_model[1]) * 0.001
			print str(scaled_model[0]), model_id, fnstr
			os.system("rosrun vfh_recognizer_db generate_views_vfh.py " + fnstr + " " + str(scaled_model[0]) + " " + str(scaled_factor) + " " + str(ITERATION))
			print str(scaled_model[0]), model_id, fnstr
