/*
 * normals_vfh_db_estimator_service.cpp
 *
 *  Created on: Jan 25, 2011
 *      Author: aaldoma
 */

#include "ros/ros.h"
#include <ros/message_traits.h>
#include <ros/serialization.h>
#include "sensor_msgs/PointCloud2.h"
#include "geometry_msgs/Pose.h"
#include "vfh_recognizer_db/normals_vfh_db_estimator.h"
#include <pcl/common/common.h>
#include "household_objects_database/database_view.h"
#include "household_objects_database/database_vfh.h"
#include "household_objects_database/database_vfh_orientation.h"
#include "household_objects_database/objects_database.h"
#include "household_objects_database/database_helper_classes.h"
#include <fstream>
#include <iostream>
#include <pcl_ros/point_cloud.h>
#include <boost/random.hpp>
#include <boost/random/normal_distribution.hpp>
#include <sys/time.h>
#include <stdio.h>
#include <unistd.h>

#include <vfh_recognizer_db/vfh_recognition.h>

vfh_recognizer_db::VFHRecognizerDB<flann::ChiSquareDistance> vfh_recognizer;

using namespace std;
namespace ser = ros::serialization;
bool USE_CLUSTER_CENTROIDS_ = true;
bool NORMALIZE_VFH_BINS_ = false;
bool FILL_SIZE_COMPONENT_ = true;
bool USE_OLD_VFH_ = false;

bool
compute_and_store (vfh_recognizer_db::normals_vfh_db_estimator::Request &req,
                   vfh_recognizer_db::normals_vfh_db_estimator::Response &res)
{
  pcl::PointCloud<pcl::PointNormal>::Ptr cloud_normals (new pcl::PointCloud<pcl::PointNormal> ());
  pcl::PointCloud<pcl::VFHSignature308>::Ptr vfh_signaturePtr (new pcl::PointCloud<pcl::VFHSignature308> ());
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud (new pcl::PointCloud<pcl::PointXYZ> ());
  pcl::PointCloud<pcl::PointXYZ>::Ptr cloud_noise (new pcl::PointCloud<pcl::PointXYZ> ());
  pcl::PointCloud<pcl::PointNormal>::Ptr cloud_normals_no_noise (new pcl::PointCloud<pcl::PointNormal> ());
  pcl::PointCloud < pcl::VFHSignature308 > vfh_signature;

  pcl::fromROSMsg (req.view, *cloud);

  //check if cloud has any points
  if (cloud->points.size () == 0)
  {
    ROS_WARN ("Cloud has no points, skip view!");
    return false;
  }

  pcl::copyPointCloud (*cloud, *cloud_noise);

  //Add noise to training data so we can use same parameters as for real data
  bool noisify = true;

  if (noisify)
  {
    double noise_std = 0.001; //1mm

    struct timeval start;
    gettimeofday (&start, NULL);

    boost::mt19937 rng;
    rng.seed (start.tv_usec);
    boost::normal_distribution<> nd (0.0, noise_std);
    boost::variate_generator<boost::mt19937&, boost::normal_distribution<> > var_nor (rng, nd);

    // Noisify each point in the dataset
    for (size_t cp = 0; cp < cloud_noise->points.size (); ++cp)
    {
      cloud_noise->points[cp].z += var_nor ();
    }
  }

  std::vector<pcl::PointCloud<pcl::VFHSignature308>,Eigen::aligned_allocator< pcl::PointCloud<pcl::VFHSignature308> > > vfh_signatures;
  std::vector < Eigen::Vector3f > centroids_dominant_orientations;

  //CONFIGURATION TO USE CLUSTERS CENTROIDS OR VIEW CENTROIDS
  vfh_recognizer.setUseClusterCentroids (USE_CLUSTER_CENTROIDS_);

  if (!NORMALIZE_VFH_BINS_)
  {
    std::cout << "BINS WONT BE NORMALIZED" << std::endl;
    vfh_recognizer.setNormalizeVFHBins (false);
  }

  vfh_recognizer.setUseOldVFH (USE_OLD_VFH_);
  vfh_recognizer.fill_size_component_ = FILL_SIZE_COMPONENT_;

  struct timeval start;
  gettimeofday (&start, NULL);

  vfh_recognizer.view_id_being_processed_ = start.tv_usec;

  vfh_recognizer.computeVFH (cloud_noise, cloud_normals, vfh_signatures, centroids_dominant_orientations);
  vfh_recognizer.computeNormals (cloud, cloud_normals_no_noise);

  vfh_signaturePtr = vfh_signatures[0].makeShared ();

  //Check the size of the normals and of the vfh_signature
  if ((cloud_normals->points.size () < 100) || (ser::serializationLength (*vfh_signaturePtr) < (sizeof(float) * 308)))
  {
    ROS_WARN ("Normals size or vfh descriptor size are wrong, skip this view - model_id: %d", req.model_id.data);
    return false;
  }

  pcl::PointCloud < pcl::PointNormal > grid;
  //grid = *cloud_normals;
  grid = *cloud_normals_no_noise; //TODO
  vfh_signature = *vfh_signaturePtr;

  //store in the database
  household_objects_database::ObjectsDatabase database ("wgs36.willowgarage.com", "5432", "willow", "willow",
                                                        "household_objects");
  if (!database.isConnected ())
  {
    std::cerr << "Database failed to connect \n";
    return -1;
  }
  std::cerr << "Database connected successfully \n";

  /* INSERT view into database */

  //compute view centroid to store it in the DB
  std::vector<float> centroid_for_db (3, 0);
  Eigen::Vector4f centroidView;
  pcl::compute3DCentroid (grid, centroidView);
  for (size_t i = 0; i < 3; i++)
    centroid_for_db[i] = centroidView[i];

  household_objects_database::DatabasePose db_pose;
  db_pose.pose_ = req.transform;
  household_objects_database::DatabaseView view_db;
  view_db.view_transform_.data () = db_pose;
  view_db.scaled_model_id_.data () = req.model_id.data;
  view_db.iteration_.data () = req.iteration.data;
  view_db.centroid_.data () = centroid_for_db;

  uint32_t msg_length = ser::serializationLength (grid);
  boost::shared_array < uint8_t > buffer (new uint8_t[msg_length]);
  ser::OStream stream (buffer.get (), msg_length);
  ser::serialize (stream, grid);

  //copy the buffer in the underlying view structure
  view_db.view_point_cloud_data_.data ().resize (msg_length);
  memcpy (&view_db.view_point_cloud_data_.data ()[0], &buffer[0], view_db.view_point_cloud_data_.data ().size ());

  if (!database.insertIntoDatabase (&view_db))
    std::cerr << "View insertion failed\n";
  else
    std::cerr << "View insertion succeeded: " << view_db.view_id_.data () << std::endl;
  ;

  if (!database.saveToDatabase (&view_db.view_point_cloud_data_))
    std::cerr << "View data insertion failed\n";

  //rename file...
  std::stringstream path_clusters_old;
  path_clusters_old << "/u/aaldoma/data/clusters/cluster_" << vfh_recognizer.view_id_being_processed_ << ".pcd";

  std::stringstream path_clusters_new;
  path_clusters_new << "/u/aaldoma/data/clusters/cluster_" << view_db.view_id_.data () << ".pcd";
  rename (path_clusters_old.str ().c_str (), path_clusters_new.str ().c_str ());

  //after view has been inserted, save to file with the view_db.view_id_.data
  std::stringstream raw_path;
  raw_path << "/u/aaldoma/data/raw_views_iteration_17/view_" << view_db.view_id_.data () << ".pcd";
  std::cout << raw_path.str () << std::endl;
  pcl::io::savePCDFileASCII (raw_path.str (), *cloud);

  for (size_t vfh_idx = 0; vfh_idx < vfh_signatures.size (); vfh_idx++)
  {

    /* INSERT VFH into database */
    household_objects_database::DatabaseVFH vfh_db;
    std::cout << "view_id:" << view_db.view_id_.data () << std::endl;
    vfh_db.view_id_.data () = view_db.view_id_.data ();
    vfh_db.iteration_.data () = view_db.iteration_.data ();

    std::vector<float> vfh_centroid (3, 0);
    for (size_t i = 0; i < 3; i++)
      vfh_centroid[i] = centroids_dominant_orientations[vfh_idx][i];

    vfh_db.centroid_.data () = vfh_centroid;

    int msg_length_vfh = ser::serializationLength (vfh_signatures[vfh_idx]);
    boost::shared_array < uint8_t > buffer_vfh (new uint8_t[msg_length_vfh]);
    ser::OStream stream_vfh (buffer_vfh.get (), msg_length_vfh);
    ser::serialize (stream_vfh, vfh_signatures[vfh_idx]);

    //copy the buffer in the underlying view structure
    vfh_db.vfh_descriptor_.data ().resize (msg_length_vfh);
    memcpy (&vfh_db.vfh_descriptor_.data ()[0], &buffer_vfh[0], vfh_db.vfh_descriptor_.data ().size ());

    if (!database.insertIntoDatabase (&vfh_db))
      std::cerr << "VFH insertion failed\n";

    if (!database.saveToDatabase (&vfh_db.vfh_descriptor_))
      std::cerr << "VFH descriptor insertion failed\n";

    if (USE_CLUSTER_CENTROIDS_)
    {
      //Once the VFH has been estimated, we need to compute the orientation histogram
      //for each VFH using the cluster centroid
      pcl::PointCloud<pcl::VFHSignature308>::Ptr
                                                 vfh_orientation_signature (
                                                                            new pcl::PointCloud<pcl::VFHSignature308> ());
      cv::Mat frequency_domain;
      vfh_recognizer.computeVFHRollOrientation (cloud_normals, vfh_orientation_signature,
                                                centroids_dominant_orientations[vfh_idx], frequency_domain); //use the centroid of the dominant clusters...

      household_objects_database::DatabaseVFHOrientation orientation_db;
      orientation_db.view_id_.data () = view_db.view_id_.data ();
      orientation_db.vfh_id_.data () = vfh_db.vfh_id_.data ();

      int msg_length_orientation = ser::serializationLength (*vfh_orientation_signature);
      boost::shared_array < uint8_t > buffer_orientation (new uint8_t[msg_length_orientation]);
      ser::OStream stream_orientation (buffer_orientation.get (), msg_length_orientation);
      ser::serialize (stream_orientation, *vfh_orientation_signature);

      //copy the buffer in the underlying view structure
      orientation_db.vfh_orientation_descriptor_.data ().resize (msg_length_orientation);
      memcpy (&orientation_db.vfh_orientation_descriptor_.data ()[0], &buffer_orientation[0],
              orientation_db.vfh_orientation_descriptor_.data ().size ());

      if (!database.insertIntoDatabase (&orientation_db))
        std::cerr << "VFH orientation insertion failed\n";

      if (!database.saveToDatabase (&orientation_db.vfh_orientation_descriptor_))
        std::cerr << "VFH orientation descriptor insertion failed\n";
    }
  }

  if (!USE_CLUSTER_CENTROIDS_)
  {
    //Once the VFH has been estimated, we need to compute the orientation histogram
    pcl::PointCloud<pcl::VFHSignature308>::Ptr vfh_orientation_signature (new pcl::PointCloud<pcl::VFHSignature308> ());
    Eigen::Vector3f ci3f (centroidView[0], centroidView[1], centroidView[2]);
    cv::Mat frequency_domain;
    vfh_recognizer.computeVFHRollOrientation (cloud_normals, vfh_orientation_signature, ci3f, frequency_domain); //use the centroid of the view

    household_objects_database::DatabaseVFHOrientation orientation_db;
    orientation_db.view_id_.data () = view_db.view_id_.data ();
    orientation_db.vfh_id_.data () = 0;

    int msg_length_orientation = ser::serializationLength (*vfh_orientation_signature);
    boost::shared_array < uint8_t > buffer_orientation (new uint8_t[msg_length_orientation]);
    ser::OStream stream_orientation (buffer_orientation.get (), msg_length_orientation);
    ser::serialize (stream_orientation, *vfh_orientation_signature);

    //copy the buffer in the underlying view structure
    orientation_db.vfh_orientation_descriptor_.data ().resize (msg_length_orientation);
    memcpy (&orientation_db.vfh_orientation_descriptor_.data ()[0], &buffer_orientation[0],
            orientation_db.vfh_orientation_descriptor_.data ().size ());

    if (!database.insertIntoDatabase (&orientation_db))
      std::cerr << "VFH orientation insertion failed\n";

    if (!database.saveToDatabase (&orientation_db.vfh_orientation_descriptor_))
      std::cerr << "VFH orientation descriptor insertion failed\n";
  }

  return true;
}

int
main (int argc, char **argv)
{

  std::cout << "CONFIGURATION:" << std::endl;
  std::cout << "USE_OLD_VFH_:" << USE_OLD_VFH_ << std::endl;
  std::cout << "USE_CLUSTER_CENTROIDS_:" << USE_CLUSTER_CENTROIDS_ << std::endl;
  std::cout << "NORMALIZE_VFH_BINS_:" << NORMALIZE_VFH_BINS_ << std::endl;
  std::cout << "FILL_SIZE_COMPONENT_:" << FILL_SIZE_COMPONENT_ << std::endl;
  ros::init (argc, argv, "normals_vfh_db_estimator_server");
  ros::NodeHandle n;

  ros::ServiceServer service = n.advertiseService ("normals_vfh_db_estimator", compute_and_store);
  ros::spin ();

  return 0;
}
