#include <vfh_recognition/vfh_recognition.h>

/*#include "household_objects_database/database_view.h"
#include "household_objects_database/database_vfh.h"
#include "household_objects_database/objects_database.h"
#include "household_objects_database/database_helper_classes.h"*/

#include <vfh_recognizer_fs/common_io.h>

//Class that implements the functionality of vfh_recognition::VFHRecognizer
//for the file system structures containing VFHS, views and so on...

namespace vfh_recognizer_fs
{
  template<template<class > class Distance>
  class VFHRecognizerFS : public vfh_recognition::VFHRecognizer<Distance>
  {

    typedef vfh_recognition::VFHRecognizer<Distance> base_t;
    using base_t::models_;
    using base_t::data_;
    using base_t::index_;

    typedef Distance<float> DistT;

    private:

      std::string root_dir;

      bool
      getPointCloudFromId (pcl::PointCloud<pcl::PointNormal> & cloud, std::string id);

      bool
      getVFHRollOrientationFromId (pcl::PointCloud<pcl::VFHSignature308> & vfh_orientation_signature, std::string id);

      bool
      getPoseFromId(std::string id, geometry_msgs::Pose & pose);

      std::string
      getModelId (std::string id);

      int
      getVFHId (std::string id);

      bool
      getVFHRollOrientationFromIdThroughView (pcl::PointCloud<pcl::VFHSignature308> & vfh_orientation_signature, std::string id);

      bool
      getVFHCentroidFromVFHid (std::vector<float> & centroid, std::string id);

      bool
      getViewCentroidFromVFHid (std::vector<float> & centroid, std::string id);

    public:

      bool
      getVFHHistogramFromVFHId (pcl::PointCloud<pcl::VFHSignature308> & vfh_descriptor, std::string vfh_id);

      ////////////////////////////////////////////////////////////////////////////////
      /** \brief Initialize the data structures for recognition
       * \param useDB - if true creates the structures from the database
       *              - if false, uses the cached kdtree if found, otherwise builds from DB
       */

      bool
      initialize (bool useDB, bf::path dpath, std::string root_dir, int linear = 0);

      ////////////////////////////////////////////////////////////////////////////////
      /** \brief Given a cluster returns the n-most similar models in models_, their poses
       * and the detection confidence.
       */

      bool
      detect (const sensor_msgs::PointCloud2ConstPtr& msg, int nModels, std::vector<std::string>& model_ids, std::vector<
          geometry_msgs::Pose> & poses, std::vector<float>& confidences, bool use_fitness_score = true);
  };

}
