/*
 * normals_vfh_db_estimator_service.cpp
 *
 *  Created on: Jan 25, 2011
 *      Author: aaldoma
 */

#include "ros/ros.h"
#include <ros/message_traits.h>
#include <ros/serialization.h>
#include "sensor_msgs/PointCloud2.h"
#include "geometry_msgs/Pose.h"
#include "vfh_recognizer_fs/remove_duplicate_views.h"
#include <vfh_recognizer_fs/vfh_recognition.h>
#include <pcl/common/common.h>
#include <fstream>
#include <iostream>
#include <pcl_ros/point_cloud.h>
#include <pcl/registration/registration.h>
#include <pcl/registration/icp.h>
#include <pcl/io/io.h>
#include <pcl/segmentation/segment_differences.h>
#include "boost/filesystem/operations.hpp"
#include "boost/filesystem/path.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/replace.hpp>
#include <boost/algorithm/string.hpp>

using namespace std;
namespace ser = ros::serialization;
namespace bf = boost::filesystem;

void
getPointCloudFromView (pcl::PointCloud<pcl::PointNormal> & cloud, std::string id, std::string root_dir) {
  stringstream view_file;
  view_file << root_dir << "/views/view_" << id << ".pcd";
  std::cout << "Going to get point cloud" << view_file.str() << std::endl;
  pcl::io::loadPCDFile(view_file.str(), cloud);
}


bool
getUniqueViews (vfh_recognizer_fs::remove_duplicate_views::Request &req,
                vfh_recognizer_fs::remove_duplicate_views::Response &res)
{

  vfh_recognizer_fs::VFHRecognizerFS < flann::ChiSquareDistance > vfh_recognizer;

  //get views from filesystem given req.model_id and build views...
  std::string MODEL_ID = req.model_id.data;
  std::string ROOT_PATH = req.root_path.data;
  stringstream view_dir_str;
  view_dir_str << ROOT_PATH << "/views";
  bf::path views_dir = view_dir_str.str();

  std::vector<std::string> views;

  bf::directory_iterator end_itr;
  for (bf::directory_iterator itr (views_dir); itr != end_itr; ++itr)
  {
    if (!bf::is_directory (itr->status ()))
    {
      std::string full_file_name = itr->path ().string ();
      //std::cerr << "opening " << full_file_name.c_str() << std::endl;
#if (!defined BOOST_FILESYSTEM_VERSION) || (BOOST_FILESYSTEM_VERSION<=2)
      std::string file_name = itr->path ().filename ();
#else
      std::string file_name = itr->path ().filename ().string ();
#endif
      std::stringstream to_search;
      to_search << "view_" << MODEL_ID << "_";
      if (!boost::algorithm::find_first(file_name,to_search.str().c_str()).empty() )
      {
        //extract view id from file_name
        std::vector<std::string> strs;
        boost::split (strs, file_name, boost::is_any_of ("_.")); //last is pcd and the one before, is the view_id

        std::stringstream to_add;
        to_add << MODEL_ID << "_" << strs[strs.size() - 2];

        std::cout << to_add.str() << std::endl;
        views.push_back(to_add.str());
      }
    }
  }

  std::cout << "views size:" << views.size() << std::endl;

  //build roll histograms
  std::vector< pcl::PointCloud<pcl::VFHSignature308>::Ptr > roll_histograms;
  roll_histograms.resize (views.size ());

  std::vector<Eigen::Vector4f,Eigen::aligned_allocator< Eigen::Vector4f> > centroids;
  centroids.resize (views.size ());
  std::vector< pcl::PointCloud < pcl::PointNormal >::Ptr > pointclouds;
  pointclouds.resize(views.size());

  for (size_t i = 0; i < views.size (); i++)
  {
    //get point cloud from view
    pcl::PointCloud < pcl::PointNormal >::Ptr cloud_normals(new pcl::PointCloud<pcl::PointNormal> ());
    getPointCloudFromView (*cloud_normals, views[i], ROOT_PATH);

    //compute centroid
    Eigen::Vector4f centroid;
    pcl::compute3DCentroid (*cloud_normals, centroid);
    Eigen::Vector3f ci3f (centroid[0], centroid[1],centroid[2]);

    //compute roll angle orientation
    pcl::PointCloud<pcl::VFHSignature308>::Ptr vfh_o_signature (new pcl::PointCloud<pcl::VFHSignature308> ());
    cv::Mat frequency_domain;
    vfh_recognizer.computeVFHRollOrientation (cloud_normals, vfh_o_signature, ci3f,frequency_domain);

    roll_histograms[i] = vfh_o_signature;
    centroids[i] = centroid;
    pointclouds[i] = cloud_normals;
  }

  //check for duplicates
  std::vector<bool> unique (views.size (), true);
  //PCLVisualizer vis ("VFH Cluster Classifier");
  //PCLVisualizer vis_diff ("Difference");

  for (size_t i = 0; i < views.size (); i++)
  {
    Eigen::Vector3f plane_normal;
    plane_normal[0] = -centroids.at(i)[0];
    plane_normal[1] = -centroids.at(i)[1];
    plane_normal[2] = -centroids.at(i)[2];
    Eigen::Vector3f z_vector = Eigen::Vector3f::UnitZ ();
    plane_normal.normalize ();
    Eigen::Vector3f axis = plane_normal.cross (z_vector);
    double rotation = -asin (axis.norm ());
    axis.normalize ();
    Eigen::Affine3f transformRM (Eigen::AngleAxisf (rotation, axis));

    pcl::PointCloud<pcl::VFHSignature308>::Ptr roll_i_ptr;
    roll_i_ptr = roll_histograms[i]; //access the correct roll histogram

    pcl::PointCloud < pcl::PointNormal >::Ptr cloud_normals_i(new pcl::PointCloud<pcl::PointNormal> ());
    cloud_normals_i = pointclouds[i];

    Eigen::Vector4f centroid_input (centroids.at(i)[0], centroids.at(i)[1], centroids.at(i)[2], 0);

    pcl::PointCloud<pcl::VFHSignature308>::Ptr vfh_o_signature_trash (new pcl::PointCloud<pcl::VFHSignature308> ());
    Eigen::Vector3f ci3f (centroids.at(i)[0], centroids.at(i)[1], centroids.at(i)[2]);
    cv::Mat frequency_domain_input;
    vfh_recognizer.computeVFHRollOrientation (cloud_normals_i, vfh_o_signature_trash, ci3f, frequency_domain_input);

    for (size_t j = (i+1); j < views.size (); j++) {
      //compare view i with the rest of the views and mark it as unique or not unique
      /*vis.removePointCloud ("cloud_i");
      vis.removePointCloud ("cloud_j");
      vis_diff.removePointCloud ("difference");*/

      pcl::PointCloud<pcl::VFHSignature308>::Ptr roll_j_ptr;
      roll_j_ptr = roll_histograms[j]; //access the correct roll histogram

      Eigen::Vector4f centroid_view (centroids.at(j)[0], centroids.at(j)[1], centroids.at(j)[2], 0);

      pcl::PointCloud < pcl::PointNormal >::Ptr cloud_normals_j(new pcl::PointCloud<pcl::PointNormal> ());
      cloud_normals_j = pointclouds[j];

      pcl::PointCloud<pcl::VFHSignature308>::Ptr vfh_o_signature_trash (new pcl::PointCloud<pcl::VFHSignature308> ());
      Eigen::Vector3f ci3f (centroids.at(j)[0], centroids.at(j)[1], centroids.at(j)[2]);
      cv::Mat frequency_domain_view;
      vfh_recognizer.computeVFHRollOrientation (cloud_normals_j, vfh_o_signature_trash, ci3f, frequency_domain_view);

      //compute roll angle!
      int angle = 0;
      angle = vfh_recognizer.computeAngleRollOrientationFrequency (frequency_domain_input, frequency_domain_view, 90, centroid_view, centroid_input , *cloud_normals_j, *cloud_normals_i, 0, *vfh_o_signature_trash, *vfh_o_signature_trash);

      //std::cout << "Computed angle:" << angle << std::endl;
      plane_normal[0] = -centroids.at(j)[0];
      plane_normal[1] = -centroids.at(j)[1];
      plane_normal[2] = -centroids.at(j)[2];
      plane_normal.normalize ();
      axis = plane_normal.cross (z_vector);
      rotation = -asin (axis.norm ());
      axis.normalize ();
      Eigen::Affine3f transformRS (Eigen::AngleAxisf (rotation, axis));
      transformRS = transformRS.inverse(); //invert transformation
      //roll transformation

      Eigen::Affine3f transformRoll(Eigen::AngleAxisf((angle*M_PI/180), Eigen::Vector3f::UnitZ()));
      Eigen::Affine3f final_trans = transformRS * transformRoll * transformRM;

      pcl::PointCloud < pcl::PointNormal > cloud_xyz;
      pcl::PointCloud < pcl::PointNormal >::Ptr gridTransformed (new pcl::PointCloud < pcl::PointNormal > ());
      pcl::transformPointCloudWithNormals(*cloud_normals_j,*gridTransformed,final_trans);
      cloud_xyz = *gridTransformed;

      // Demean the cloud
      pcl::PointCloud < pcl::PointNormal >::Ptr cloud_xyz_demean(new pcl::PointCloud<pcl::PointNormal> ());
      Eigen::Vector4f diff = Eigen::Vector4f::Zero();

      Eigen::Vector3f centr(centroids.at(j)[0],centroids.at(j)[1],centroids.at(j)[2]);
      centr = final_trans * centr;

      diff[0] = -centroids.at(i)[0] + centr[0];
      diff[1] = -centroids.at(i)[1] + centr[1];
      diff[2] = -centroids.at(i)[2] + centr[2];

      pcl::demeanPointCloud (cloud_xyz, diff, *cloud_xyz_demean);

      //we can compare cloud_xyz_demean with cloud_normals_i
      pcl::IterativeClosestPoint < pcl::PointNormal, pcl::PointNormal > reg;
      reg.setInputCloud (cloud_xyz_demean);
      reg.setInputTarget (cloud_normals_i);
      double fscore = reg.getFitnessScore (0.005);

      //segment difference
      pcl::SegmentDifferences<pcl::PointNormal> seg;
      typedef pcl::KdTree<pcl::PointNormal>::Ptr KdTreePtr;
      KdTreePtr normals_tree = boost::make_shared<pcl::KdTreeFLANN<pcl::PointNormal> > (true);
      //normals_tree->setInputCloud (cloud_xyz_demean);
      seg.setDistanceThreshold (0.01*0.01); //1cm
      seg.setSearchMethod(normals_tree);
      seg.setInputCloud(cloud_normals_i);
      seg.setTargetCloud(cloud_xyz_demean);

      pcl::PointCloud < pcl::PointNormal > difference;
      seg.segment(difference);

      //std::cout << "Fitness score:" << fscore << std::endl;
      fscore = (double)(difference.points.size()) / (double)(cloud_normals_i->points.size());
      std::cout << "% of remaining points:" << fscore << " i:" << i << " j:" << j << " size diff:" << difference.points.size() << std::endl;
      if (fscore < 0.025) { //if less than 2.5% of the points have their nearest neighbor outside threshold, then view is not unique
        //std::cout << "View is not unique" << std::endl;
        unique[i] = false;
        break; //we have found at least one view after i that is the same
      }
    }
  }

  //remove duplicate views and descriptors from the FS

  int unique_views = 0;
  for (size_t i = 0; i < unique.size (); i++)
  {
    if (!unique.at (i)) //if its not unique, delete it from database
    {

      std::string view_id = views.at(i);
      std::cout << "View "<< view_id << " not unique" << std::endl;

      //delete view
      stringstream view_dir_str;
      view_dir_str << ROOT_PATH << "/views/view_" << view_id << ".pcd";
      bf::path view_path = view_dir_str.str();

      bf::remove(view_path);

      //get VFHid mapping from this view
      stringstream vfh_mappings;
      vfh_mappings << ROOT_PATH << "/views_vfh_map/view_vfh_" << view_id << ".txt";
      std::ifstream in;
      in.open (vfh_mappings.str ().c_str (), std::ifstream::in);

      char linebuf[256];
      in.getline (linebuf, 256);
      std::string line (linebuf);

      std::vector < std::string > strs;
      boost::split (strs, line, boost::is_any_of (" "));

      for(size_t k=0; k < strs.size(); k++) { //maybe size - 1
        //delete transforms,vfhs,vfhs_roll_orientation,centroids, vfh_view_map
        std::string vfh_id = strs[k];

        stringstream vfh_path;
        vfh_path << ROOT_PATH << "/vfhs/vfh_" << MODEL_ID << "_" << vfh_id << ".pcd";
        stringstream vfh_roll_path;
        vfh_roll_path << ROOT_PATH << "/vfhs_roll_orientation/vfh_orientation_" << MODEL_ID << "_" << vfh_id << ".pcd";
        stringstream transform_path;
        transform_path << ROOT_PATH << "/transforms/transform_" << MODEL_ID << "_" << vfh_id << ".txt";
        stringstream centroid_path;
        centroid_path << ROOT_PATH << "/centroids/centroid_" << MODEL_ID << "_" << vfh_id << ".txt";
        stringstream vfh_view_map_path;
        vfh_view_map_path << ROOT_PATH << "/vfh_views_map/vfh_view_" << MODEL_ID << "_" << vfh_id << ".txt";

        bf::remove(bf::path(vfh_path.str()));
        bf::remove(bf::path(vfh_roll_path.str()));
        bf::remove(bf::path(transform_path.str()));
        bf::remove(bf::path(centroid_path.str()));
        bf::remove(bf::path(vfh_view_map_path.str()));
      }

      //delete view vfh map
      bf::remove(bf::path(vfh_mappings.str()));
    } else {
      unique_views++;
    }
  }

  std::cout << "Number of unique views:" << unique_views << std::endl;

  return true;
}

int
main (int argc, char **argv)
{
  ros::init (argc, argv, "remove_duplicate_views_server");
  ros::NodeHandle n;

  ros::ServiceServer service = n.advertiseService ("remove_duplicate_views", getUniqueViews);
  ros::spin ();

  return 0;
}
